""" This is the mogo syntactic sugar library for MongoDB. """

from mogo.model import Model, PolyModel
from mogo.field import Field, ReferenceField, ConstantField, EnumField
from mogo.cursor import ASC, DESC
from mogo.connection import connect, session

# Allows flexible (probably dangerous) automatic field creation for
# /really/ schemaless designs.
AUTO_CREATE_FIELDS = False

__all__ = [
    'Model', 'PolyModel', 'Field', 'ReferenceField', "ConstantField",
    "EnumField", 'connect', 'session', 'ASC', 'DESC', "AUTO_CREATE_FIELDS"
]



""" The wrapper for pymongo's connection stuff. """

import urlparse
from pymongo import MongoClient
from pymongo.errors import ConnectionFailure


class Connection(object):
    """
    This just caches a pymongo connection and adds
    a few shortcuts.
    """

    _instance = None
    connection = None
    _database = None

    @classmethod
    def instance(cls):
        """ Retrieves the shared connection. """
        if not cls._instance:
            cls._instance = Connection()
        return cls._instance

    @classmethod
    def connect(cls, database=None, uri="mongodb://localhost:27017", **kwargs):
        """
        Wraps a pymongo connection.
        TODO: Allow some of the URI stuff.
        """
        if not database:
            database = urlparse.urlparse(uri).path
            while database.startswith("/"):
                database = database[1:]
            if not database:
                raise ValueError("A database name is required to connect.")

        conn = cls.instance()
        conn._database = database
        conn.connection = MongoClient(uri, **kwargs)
        return conn.connection

    def get_database(self, database=None):
        """ Retrieves a database from an existing connection. """
        if not self.connection:
            raise ConnectionFailure('No connection')
        if not database:
            if not self._database:
                raise Exception('No database submitted')
            database = self._database
        return self.connection[database]

    def get_collection(self, collection, database=None):
        """ Retrieve a collection from an existing connection. """
        return self.get_database(database=database)[collection]


class Session(object):
    """ This class just wraps a connection instance """

    def __init__(self, database, *args, **kwargs):
        """ Stores a connection instance """
        self.connection = None
        self.database = database
        self.args = args
        self.kwargs = kwargs

    def connect(self):
        """ Connect to MongoDB """
        connection = Connection()
        connection._database = self.database
        connection.connection = MongoClient(*self.args, **self.kwargs)
        self.connection = connection

    def disconnect(self):
        # PyMongo removed the disconnect keyword, close() is now used.
        self.close()

    def close(self):
        self.connection.connection.close()

    def __enter__(self):
        """ Open the connection """
        self.connect()
        return self

    def __exit__(self, type, value, traceback):
        """ Close the connection """
        self.disconnect()


def connect(*args, **kwargs):
    """
    Initializes a connection and the database. It returns
    the pymongo connection object so that end_request, etc.
    can be called if necessary.
    """
    return Connection.connect(*args, **kwargs)


def session(database, *args, **kwargs):
    """
    Returns a session object to be used with the `with` statement.
    """
    return Session(database, *args, **kwargs)



"""
Really, really basic around pymongo.Cursor. Just makes sure
that a result dict is wrapped in a Model to keep everything
clean.
"""

from pymongo.cursor import Cursor as PyCursor
from pymongo import ASCENDING, DESCENDING

# Shortcuts are better! :)
ASC = ASCENDING
DESC = DESCENDING


class Cursor(PyCursor):
    """ A simple wrapper around pymongo's Cursor class. """

    def __init__(self, model, spec=None, *args, **kwargs):
        self._order_entries = []
        self._query = spec
        self._model = model
        PyCursor.__init__(
            self, model._get_collection(), spec, *args, **kwargs)

    def next(self):
        value = PyCursor.next(self)
        return self._model(**value)

    # convenient because if it quacks like a list...
    def __len__(self):
        return self.count()

    def __getitem__(self, *args, **kwargs):
        value = PyCursor.__getitem__(self, *args, **kwargs)
        if type(value) == self.__class__:
            return value
        return self._model(**value)

    def first(self):
        if self.count() == 0:
            return None
        return self[0]

    def order(self, **kwargs):
        if len(kwargs) != 1:
            raise ValueError("order() requires one field = ASC or DESC.")
        for key, value in kwargs.iteritems():
            if value not in (ASC, DESC):
                raise TypeError("Order value must be mogo.ASC or mogo.DESC.")
            self._order_entries.append((key, value))
            # According to the docs, only the LAST .sort() matters to
            # pymongo, so this SHOULD be safe
            self.sort(self._order_entries)
        return self

    def update(self, modifier):
        if self._query is None:
            raise ValueError(
                "Cannot update on a cursor without a query. If you "
                "actually want to modify all values on a model, pass "
                "in an explicit {} to find().")
        self._model.update(self._query, modifier, multi=True)
        return self

    def change(self, **kwargs):
        modifier = {"$set": kwargs}
        return self.update(modifier)



""" The basic field attributes. """

import warnings

# PyMongo moving libraries around -- eventually this can
# go away.
try:
    from pymongo.dbref import DBRef
except ImportError:
    from bson.dbref import DBRef


class EmptyRequiredField(Exception):
    """ Raised when a required field is not set on the model instance. """
    pass


class Field(object):
    """
    This class is responsible for type-checking, set- and get- callbacks,
    coercing values, and general data-y things. More simply, it is responsible
    for storing things into mongo-friendly values and retrieving them in
    application-friendly values.
    """

    value_type = None

    def __init__(self, value_type=None, **kwargs):
        self.value_type = value_type or self.value_type
        self.required = kwargs.get("required", False) is True
        if "default" in kwargs:
            self.default = kwargs["default"]
        set_callback = getattr(self, "_set_callback", None)
        get_callback = getattr(self, "_get_callback", None)
        coerce_callback = getattr(self, "_coerce_callback", None)
        self._set_callback = kwargs.get("set_callback", set_callback)
        self._get_callback = kwargs.get("get_callback", get_callback)
        self._coerce_callback = kwargs.get("coerce_callback", coerce_callback)
        self.id = id(self)
        self._field_name = kwargs.get("field_name", None)

    def __get__(self, instance, klass=None):
        if instance is None:
            # Classes see the descriptor itself
            return self
        value = self._get_value(instance)
        return value

    def _get_field_name(self, model_instance):
        """ Try to retrieve field name from instance """
        if self._field_name:
            return self._field_name
        fields = getattr(model_instance, "_fields")
        return fields[self.id]

    def _get_value(self, instance):
        """ Retrieve the value from the instance """
        field_name = self._get_field_name(instance)
        value = None
        if not field_name in instance:
            if self.required:
                raise EmptyRequiredField(
                    "'%s' is required but is empty." % field_name)
            self._set_default(instance, field_name)
        value = instance.get(field_name)
        if self._get_callback:
            value = self._get_callback(instance, value)
        return value

    def _set_default(self, model, field):
        if field in model:
            # value already set, not overwriting it.
            return
        if hasattr(self, "default"):
            if not callable(self.default):
                default_value = self.default
            else:
                default_value = self.default()
            setattr(model, field, default_value)

    def _check_value_type(self, value, field_name):
        """ Verifies that a value is the proper type """
        if value is not None and self.value_type is not None:
            valid = isinstance(value, self.value_type)
            if not valid:
                value_type = type(value)
                raise TypeError(
                    "Invalid type %s instead of %s for field '%s'" % (
                        value_type, self.value_type, field_name))

    def __set__(self, instance, value):
        field_name = self._get_field_name(instance)
        try:
            self._check_value_type(value, field_name)
        except TypeError:
            if not self._coerce_callback:
                raise
            value = self._coerce_callback(value)
            self._check_value_type(value, field_name)

        if self._set_callback:
            value = self._set_callback(instance, value)
        instance[field_name] = value


class ReferenceField(Field):
    """ Simply holds information about the reference model. """

    def __init__(self, model, **kwargs):
        super(ReferenceField, self).__init__(model, **kwargs)
        self.model = model

    def _set_callback(self, instance, value):
        """ Resolves a Model to a DBRef """
        if value:
            value = DBRef(self.model._get_name(), value.id)
        return value

    def _get_callback(self, instance, value):
        """ Retrieves the id, then retrieves the model from the db """
        if value:
            # Should be a DBRef
            return self.model.find_one({"_id": value.id})
        return value


class ConstantField(Field):
    """ Doesn't let you change the value after setting it. """

    def _set_callback(self, instance, value):
        """ Block changing values from being set. """
        if instance._get_id() and value is not self._get_value(instance):
            raise ValueError("Constant fields cannot be altered after saving.")
        return value


class EnumField(Field):
    """ Only accepts values from a set / list of values.
    The first argument should be an iterable with acceptable values, or
    optionally a callable that takes the instance as the first argument and
    returns an iterable with acceptable values.

    For instance, both of these are valid:

        EnumField(("a", "b", 5))
        EnumField(lambda x: [5, 6])

    """

    def __init__(self, iterable, **kwargs):
        super(EnumField, self).__init__(**kwargs)
        self.iterable = iterable

    def _set_callback(self, instance, value):
        """ Checks for value in iterable. """
        accepted_values = self.iterable
        if callable(self.iterable):
            accepted_values = self.iterable(instance)
        if value not in accepted_values:
            # not listing the accepted values because that might be bad,
            # for example, if it's a cursor or other exhaustible iterator
            raise ValueError("Value %s not in acceptable values." % value)
        return value
